# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np
import xgboost
from sklearn.compose import ColumnTransformer
from sklearn.feature_extraction import DictVectorizer
from sklearn.multiclass import OneVsRestClassifier
from sklearn.pipeline import Pipeline

from bugbug import bug_features, bugzilla, feature_cleanup, utils
from bugbug.model import BugModel

KEYWORD_DICT = {
    "sec-critical": "security",
    "sec-high": "security",
    "sec-moderate": "security",
    "sec-low": "security",
    "sec-other": "security",
    "sec-audit": "security",
    "sec-vector": "security",
    "sec-want": "security",
    "memory-footprint": "memory",
    "memory-leak": "memory",
    "crash": "crash",
    "crashreportid": "crash",
    "perf": "performance",
}
KEYWORD_LIST = sorted(set(KEYWORD_DICT.values()))


class BugTypeModel(BugModel):
    def __init__(self, lemmatization=False, historical=False):
        BugModel.__init__(self, lemmatization)

        self.calculate_importance = False

        feature_extractors = [
            bug_features.has_str(),
            bug_features.severity(),
            # Ignore keywords that would make the ML completely skewed
            # (we are going to use them as 100% rules in the evaluation phase).
            bug_features.keywords(set(KEYWORD_DICT.keys())),
            bug_features.is_coverity_issue(),
            bug_features.has_crash_signature(),
            bug_features.has_url(),
            bug_features.has_w3c_url(),
            bug_features.has_github_url(),
            bug_features.whiteboard(),
            bug_features.patches(),
            bug_features.landings(),
            bug_features.blocked_bugs_number(),
            bug_features.ever_affected(),
            bug_features.affected_then_unaffected(),
            bug_features.product(),
            bug_features.component(),
        ]

        cleanup_functions = [
            feature_cleanup.url(),
            feature_cleanup.fileref(),
            feature_cleanup.synonyms(),
        ]

        self.extraction_pipeline = Pipeline(
            [
                (
                    "bug_extractor",
                    bug_features.BugExtractor(feature_extractors, cleanup_functions),
                ),
                (
                    "union",
                    ColumnTransformer(
                        [
                            ("data", DictVectorizer(), "data"),
                            ("title", self.text_vectorizer(min_df=0.001), "title"),
                            (
                                "first_comment",
                                self.text_vectorizer(min_df=0.001),
                                "first_comment",
                            ),
                            (
                                "comments",
                                self.text_vectorizer(min_df=0.001),
                                "comments",
                            ),
                        ]
                    ),
                ),
            ]
        )

        self.clf = OneVsRestClassifier(
            xgboost.XGBClassifier(n_jobs=utils.get_physical_cpu_count())
        )

    def get_labels(self):
        classes = {}

        for bug_data in bugzilla.get_bugs():
            target = np.zeros(len(KEYWORD_LIST))
            for keyword in bug_data["keywords"]:
                if keyword in KEYWORD_DICT:
                    target[KEYWORD_LIST.index(KEYWORD_DICT[keyword])] = 1

            classes[int(bug_data["id"])] = target

        return classes, KEYWORD_LIST

    def get_feature_names(self):
        return self.extraction_pipeline.named_steps["union"].get_feature_names()

    def overwrite_classes(self, bugs, classes, probabilities):
        for i, bug in enumerate(bugs):
            for keyword in bug["keywords"]:
                if keyword in KEYWORD_LIST:
                    if probabilities:
                        classes[i][KEYWORD_LIST.index(keyword)] = 1.0
                    else:
                        classes[i][KEYWORD_LIST.index(keyword)] = 1

        return classes



# -*- coding: utf-8 -*-

import argparse
import sys
from logging import INFO, basicConfig, getLogger

from bugbug.utils import download_and_load_model

basicConfig(level=INFO)
logger = getLogger(__name__)


class ModelChecker:
    def go(self, model_name):
        # Load the model
        model = download_and_load_model(model_name)

        # Then call the check method of the model
        success = model.check()

        if not success:
            msg = f"Check of model {model.__class__!r} failed, check the output for reasons why"
            logger.warning(msg)
            sys.exit(1)


def main():
    description = "Check the models"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("model", help="Which model to check.")

    args = parser.parse_args()

    checker = ModelChecker()
    checker.go(args.model)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import os

from setuptools import find_packages, setup

here = os.path.dirname(__file__)


def read_requirements(file_):
    with open(os.path.join(here, file_)) as f:
        return sorted(list(set(line.split("#")[0].strip() for line in f)))


install_requires = read_requirements("requirements.txt")


with open(os.path.join(here, "VERSION")) as f:
    version = f.read().strip()

# Read the extra requirements
extras = ["nlp", "nn"]

extras_require = {}

for extra in extras:
    extras_require[extra] = read_requirements("extra-%s-requirements.txt" % extra)


setup(
    name="bugbug",
    version=version,
    description="ML tools for Mozilla projects",
    author="Marco Castelluccio",
    author_email="mcastelluccio@mozilla.com",
    install_requires=install_requires,
    extras_require=extras_require,
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    include_package_data=True,
    license="MPL2",
    entry_points={
        "console_scripts": [
            "bugbug-data-commits = scripts.commit_retriever:main",
            "bugbug-data-bugzilla = scripts.bug_retriever:main",
            "bugbug-data-test-scheduling-history = scripts.test_scheduling_history_retriever:main",
            "bugbug-data-revisions = scripts.revision_retriever:main",
            "bugbug-train = scripts.trainer:main",
            "bugbug-train-similarity = scripts.similarity_trainer:main",
            "bugbug-check = scripts.check:main",
            "bugbug-microannotate-generate = scripts.microannotate_generator:main",
            "bugbug-classify-commit = scripts.commit_classifier:main",
            "bugbug-classify-bug = scripts.bug_classifier:main",
            "bugbug-regressor-finder = scripts.regressor_finder:main",
            "bugbug-retrieve-training-metrics = scripts.retrieve_training_metrics:main",
            "bugbug-analyze-training-metrics = scripts.analyze_training_metrics:main",
            "bugbug-check-all-metrics = scripts.check_all_metrics:main",
            "bugbug-past-bugs-by-unit = scripts.past_bugs_by_unit:main",
            "bugbug-testing-policy-stats = scripts.testing_policy_stats:main",
            "bugbug-generate-landings-risk-report = scripts.generate_landings_risk_report:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
    ],
)



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
from collections import defaultdict
from functools import lru_cache

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import TfidfVectorizer

HAS_OPTIONAL_DEPENDENCIES = False

try:
    import spacy
    from gensim.models import KeyedVectors
    from spacy.tokenizer import Tokenizer

    HAS_OPTIONAL_DEPENDENCIES = True
except ImportError:
    pass

try:
    if HAS_OPTIONAL_DEPENDENCIES:
        nlp = spacy.load("en_core_web_sm")
except OSError:
    msg = (
        "Spacy model is missing, install it with: "
        f"{sys.executable} -m spacy download en_core_web_sm"
    )
    print(msg, file=sys.stderr)

OPT_MSG_MISSING = (
    "Optional dependencies are missing, install them with: pip install bugbug[nlp]\n"
    "You might need also to download the models with: "
    f"{sys.executable} -m spacy download en_core_web_sm"
)


def spacy_token_lemmatizer(text):
    if len(text) > nlp.max_length:
        text = text[: nlp.max_length - 1]
    doc = nlp(text)
    return [token.lemma_ for token in doc]


class SpacyVectorizer(TfidfVectorizer):
    def __init__(self, *args, **kwargs):

        # Detect when the Spacy optional dependency is missing
        if not HAS_OPTIONAL_DEPENDENCIES:
            raise NotImplementedError(OPT_MSG_MISSING)

        super().__init__(tokenizer=spacy_token_lemmatizer, *args, **kwargs)


@lru_cache()
def get_word_embeddings():
    word_embeddings = KeyedVectors.load_word2vec_format("wiki-news-300d-1M-subword.vec")
    word_embeddings.init_sims(replace=True)
    return word_embeddings


class MeanEmbeddingTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        # Detect when the Gensim optional dependency are missing
        if not HAS_OPTIONAL_DEPENDENCIES:
            raise NotImplementedError(OPT_MSG_MISSING)

        self.model = get_word_embeddings()
        self.dim = len(self.model["if"])

    def fit(self, x, y=None):
        return self

    def transform(self, data):
        tokenizer = Tokenizer(nlp.vocab)
        return np.array(
            [
                np.mean(
                    [
                        self.model[w.text.lower()]
                        for w in words
                        if w.text.lower() in self.model
                    ]
                    or [np.zeros(self.dim)],
                    axis=0,
                )
                for words in tokenizer.pipe(data)
            ]
        )

    def get_feature_names(self):
        return np.array([f"_{i}" for i in range(self.dim)], dtype=object)


class TfidfMeanEmbeddingTransformer(MeanEmbeddingTransformer):
    def __init__(self):
        super().__init__()
        self.word2weight = None

    def fit(self, X, y=None):
        tfidf = TfidfVectorizer(analyzer=lambda x: x)
        tfidf.fit(X)

        # If a word was never seen, it must be at least as infrequent as any of the known words.
        # So, the default idf is the max of known idfs.
        max_idf = max(tfidf.idf_)
        self.word2weight = defaultdict(
            lambda: max_idf, [(w, tfidf.idf_[i]) for w, i in tfidf.vocabulary_.items()]
        )

        return self

    def transform(self, data):
        tokenizer = Tokenizer(nlp.vocab)
        return np.array(
            [
                np.mean(
                    [
                        self.model[w.text.lower()] * self.word2weight[w.text.lower()]
                        for w in words
                        if w.text.lower() in self.model
                    ]
                    or [np.zeros(self.dim)],
                    axis=0,
                )
                for words in tokenizer.pipe(data)
            ]
        )



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import dateutil.parser
from dateutil.relativedelta import relativedelta

from bugbug import bugzilla


def bool_str(val):
    assert val in ["", "0", "1"], f"Unexpected boolean value: '{val}'"

    return True if val == "1" else False


def keyword_mapping(keyword):
    mapping = {
        "mlk": "memory-leak",
        "topmlk": "top-memory-leak",
        "pp": "platform-parity",
        "footprint": "memory-footprint",
        "ateam-marionette-firefox-puppeteer": "pi-marionette-firefox-puppeteer",
        "ateam-marionette-big": "pi-marionette-big",
        "ateam-marionette-runner": "pi-marionette-runner",
        "ateam-marionette-server": "pi-marionette-server",
        "ateam-marionette-client": "pi-marionette-client",
        "ateam-marionette-intermittent": "pi-marionette-intermittent",
        "ateam-marionette-harness-tests": "pi-marionette-harness-tests",
        "ateam-marionette-spec": "pi-marionette-spec",
        "csec-dos": "csectype-dos",
        "csec-oom": "csectype-oom",
        "csec-bounds": "csectype-bounds",
        "bug-quality": "bmo-bug-quality",
    }

    return mapping[keyword] if keyword in mapping else keyword


def group_mapping(group):
    mapping = {"release-core-security": "core-security-release"}

    return mapping[group] if group in mapping else group


def cf_rank(val):
    if val in ["", "0"]:
        return None

    return val


# E.g. https://bugzilla.mozilla.org/rest/bug/1162372.
def version_to_branch(version):
    if version.startswith("Firefox "):
        return f'{version[len("Firefox "):]} Branch'

    return version


def op_sys(op_sys):
    if op_sys == "Mac OS X":
        return "macOS"

    return op_sys


def platform(platform):
    if platform == "Macintosh":
        return "PowerPC"
    elif platform == "PC":
        return "x86"

    return platform


def product(product):
    mapping = {
        "Web Compatibility Tools": "Web Compatibility",
        "Mozilla Developer Network": "developer.mozilla.org",
        "MozReview": "MozReview Graveyard",
        "mozilla.org graveyard": "mozilla.org Graveyard",
        "TaskCluster": "Taskcluster",
        "Firefox OS": "Firefox OS Graveyard",
        "Add-on SDK": "Add-on SDK Graveyard",
        "Connected Devices": "Connected Devices Graveyard",
        "Seamonkey": "SeaMonkey",
        "Mozilla Application Suite": "SeaMonkey",
        "Mozilla Services": "Cloud Services",
        "Browser": "Core",
    }

    return mapping[product] if product in mapping else product


def target_milestone(target_milestone):
    if target_milestone.startswith("Seamonkey"):
        return target_milestone.lower()

    mapping = {"6.2.2": "6.2.2.1", "Firefox 3.7": "Firefox 4.0"}

    return (
        mapping[target_milestone] if target_milestone in mapping else target_milestone
    )


def null_str(val):
    if val == "":
        return None

    return val


FIELD_TYPES = {
    "blocks": int,
    "depends_on": int,
    "regressed_by": int,
    "regressions": int,
    "is_confirmed": bool_str,
    "is_cc_accessible": bool_str,
    "is_creator_accessible": bool_str,
    "cf_rank": cf_rank,
    "keywords": keyword_mapping,
    "groups": group_mapping,
    "op_sys": op_sys,
    "platform": platform,
    "product": product,
    "target_milestone": target_milestone,
    "cf_due_date": null_str,
}


def is_email(val):
    return isinstance(val, str) and "@" in val


def is_expected_inconsistent_field(field, last_product, bug_id):
    # TODO: Remove the Graveyard case when https://bugzilla.mozilla.org/show_bug.cgi?id=1541926 is fixed.
    return (
        (field.startswith("cf_") and last_product == "Firefox for Android Graveyard")
        or (field == "cf_status_firefox_esr52" and bug_id == 1280099)
        or (
            field == "cf_status_firefox57"
            and bug_id
            in (1382577, 1382605, 1382606, 1382607, 1382609, 1394996, 1406290, 1407347)
        )
        or (field == "cf_status_firefox58" and bug_id in {1280099, 1328936, 1394996})
        or (field == "cf_status_firefox59" and bug_id in {1280099, 1328936, 1394996})
        or (
            field == "cf_tracking_firefox59"
            and bug_id in (1328936, 1394996, 1_443_367, 1_443_630)
        )
        or (
            field == "cf_status_firefox60"
            and bug_id
            in [
                1_442_627,
                1_443_505,
                1_443_599,
                1_443_600,
                1_443_603,
                1_443_605,
                1_443_608,
                1_443_609,
                1_443_611,
                1_443_614,
                1_443_615,
                1_443_617,
                1_443_644,
                1443466,
            ]
        )
        or (field in ["cf_has_str", "cf_has_regression_range"] and bug_id == 1_440_338)
        or (field == "cf_has_regression_range" and bug_id == 1542185)
        or (
            field == "cf_has_str" and bug_id == 1462571
        )  # TODO: Remove when https://bugzilla.mozilla.org/show_bug.cgi?id=1550104 is fixed
    )


def is_expected_inconsistent_change_field(field, bug_id, new_value, new_value_exp):
    # The 'enhancement' severity has been removed, but it doesn't show up in the history.
    # See https://bugzilla.mozilla.org/show_bug.cgi?id=1541362.
    return (
        (field in ["status", "resolution", "cf_last_resolved"] and bug_id == 1_312_722)
        or (field == "cf_last_resolved" and bug_id == 1_321_567)
        or (
            field == "url"
            and bug_id
            in (
                380637,
                740_223,
                1326518,
                1335350,
                1340490,
                1378065,
                1381475,
                1389540,
                1395484,
                1403353,
            )
        )
        or (field == "severity" and new_value == "enhancement")
        or (field == "cf_blocking_20" and bug_id == 380637)
        or (field == "cf_blocking_191" and bug_id in (471015, 607222))
        or (field == "cf_blocking_192" and bug_id == 607222)
        or (field == "cf_status_firefox_esr45" and bug_id == 1292534)
        or (
            field == "cf_status_firefox_esr52"
            and bug_id in [1_436_341, 1_443_518, 1_443_637]
        )
        or (
            field == "cf_status_firefox57"
            and bug_id
            in [
                1_328_936,
                1_381_197,
                1_382_577,
                1_382_605,
                1_382_606,
                1_382_607,
                1_382_609,
                1_383_711,
                1_387_511,
                1_394_996,
                1_403_927,
                1_403_977,
                1_404_917,
                1_406_290,
                1_407_347,
                1_409_651,
                1_410_351,
            ]
        )
        or (
            field == "cf_status_firefox58"
            and bug_id
            in [
                1_328_936,
                1_383_870,
                1_394_996,
                1_397_772,
                1_408_468,
                1_418_410,
                1_436_341,
                1_441_537,
                1_443_511,
                1_443_518,
                1_443_527,
                1_443_544,
                1_443_612,
                1_443_630,
                1_443_637,
            ]
        )
        or (
            field == "cf_status_firefox59"
            and bug_id
            in [
                1_328_936,
                1_394_996,
                1_397_772,
                1_403_334,
                1_428_996,
                1_431_306,
                1_436_341,
                1_441_537,
                1_443_511,
                1_443_518,
                1_443_527,
                1_443_533,
                1_443_544,
                1_443_612,
                1_443_630,
                1_443_637,
            ]
        )
        or (
            field == "cf_status_firefox60"
            and bug_id
            in [
                851471,
                1280099,
                1_362_303,
                1_363_862,
                1_375_913,
                1_390_583,
                1_401_847,
                1_402_845,
                1_414_901,
                1_421_387,
                1_434_483,
                1_434_869,
                1_436_287,
                1_437_803,
                1_438_608,
                1_440_146,
                1_441_052,
                1_442_160,
                1_442_186,
                1_442_861,
                1_443_205,
                1_443_368,
                1_443_371,
                1_443_438,
                1_443_507,
                1_443_511,
                1_443_518,
                1_443_525,
                1_443_527,
                1_443_528,
                1_443_533,
                1_443_560,
                1_443_578,
                1_443_585,
                1_443_593,
                1_443_612,
                1_443_630,
                1_443_637,
                1_443_646,
                1_443_650,
                1_443_651,
                1_443_664,
                1442543,
                1443180,
                1443446,
                1443522,
            ]
        )
        or (field == "cf_tracking_firefox60" and bug_id in [1_375_913, 1_439_875])
        or (field == "priority" and bug_id in {1215089, 1_337_747})
        or (
            field == "type" and bug_id == 1540796
        )  # TODO: Remove once https://bugzilla.mozilla.org/show_bug.cgi?id=1550120 is fixed.
        or (
            field == "cf_last_resolved" and bug_id == 1540998
        )  # TODO: Remove once https://bugzilla.mozilla.org/show_bug.cgi?id=1550128 is fixed.
        or (
            field == "type" and bug_id == 1257155
        )  # TODO: Remove once https://bugzilla.mozilla.org/show_bug.cgi?id=1550129 is fixed.
        or (bug_id == 1370035 and field in ("cf_has_str", "cf_has_regression_range"))
        or (bug_id == 1400540 and field in ("target_milestone", "status", "resolution"))
        or (bug_id == 1402929 and field == "priority")
        or (field == "whiteboard" and new_value.rstrip() == new_value_exp.rstrip())
        or (
            field == "summary"
            and bug_id
            in (
                1326589,
                1350800,
                1368617,
                1396657,
                1399203,
                1405388,
                1405496,
                1438422,
                1440635,
                1447653,
                1452217,
                1453584,
                1462986,
                1467331,
                1478399,
                1482142,
                1495267,
                1500185,
                1510849,
                1531130,
                1578734,
                1591051,
            )
        )  # https://bugzilla.mozilla.org/show_bug.cgi?id=1556319
        or (field == "whiteboard" and bug_id in (1385923, 1340867))
        or (
            field == "url"
            and bug_id
            in (1362789, 1364792, 1431604, 1437528, 1445898, 1446685, 1460828, 1494587)
        )
        or (field in ("platform", "op_sys") and bug_id == 568516)
        or (
            field == "target_milestone"
            and bug_id in {11050, 19462, 106327, 107264, 144795, 306730}
        )
        or (field == "product" and bug_id in {21438, 263013})
        or is_email(
            new_value
        )  # TODO: Users can change their email, try with all emails from a mapping file.
    )


def is_expected_inconsistent_change_list_field(field, bug_id, value):
    return (
        (
            field == "keywords"
            and value == "checkin-needed"
            and bug_id in [1274602, 1341727]
        )
        or (
            field == "keywords"
            and value
            in [
                "patch",
                "nsbeta1",
                "nsbeta1+",
                "nsbeta1-",
                "nsbeta3",
                "mozilla1.1",
                "mozilla1.0",
                "4xp",
                "sec-review-complete",
                "beta1",
                "beta2",
                "mozilla0.9",
                "mozilla0.9.2",
                "verified1.0.1",
                "fixed1.9.0.10",
                "adt1.0.1+",
                "mail3",
            ]
        )  # These keywords don't exist anymore.
        or is_email(
            value
        )  # TODO: Users can change their email, try with all emails from a mapping file.
    )


def is_expected_inconsistent_change_flag(flag, obj_id):
    # TODO: always assert here, once https://bugzilla.mozilla.org/show_bug.cgi?id=1514415 is fixed.
    return (
        obj_id in [1_052_536, 1_201_115, 1_213_517, 794_863]
        or (
            flag == "in-testsuite+"
            and obj_id
            in [1_318_438, 1_312_852, 1_332_255, 1_344_690, 1_362_387, 1_380_306]
        )
        or (flag == "in-testsuite-" and obj_id in {906177, 1321444, 1342431, 1370129})
        or (
            flag == "checkin+"
            and obj_id
            in {
                8795236,
                8795632,
                8791622,
                8794855,
                8795623,
                8792801,
                8791937,
                8795246,
                8795282,
                8786330,
                8786345,
                8787093,
                8795228,
                8795333,
                8880381,
                8879995,
                8872652,
                8871000,
                8870452,
                8870505,
                8864140,
                8868787,
            }
        )
        or (flag == "checkin-" and obj_id == 8924974)
        or (
            flag == "webcompat?"
            and obj_id
            in (
                1360579,
                1326028,
                1356114,
                1360238,
                1364598,
                1367657,
                1375319,
                1382724,
                1397981,
                1401593,
                1405744,
                1416728,
                1417293,
                1428263,
                1469747,
                1522872,
                1531758,
            )
        )
        or (
            flag == "webcompat+"
            and obj_id in (1294490, 1443958, 1455894, 1456313, 1489308)
        )
        or (flag == "webcompat-" and obj_id == 1419848)
        or (flag == "qe-verify+" and obj_id in {1567624, 1572197})
        or (
            flag == "qe-verify-"
            and obj_id
            in [
                1282408,
                1322685,
                1336510,
                1363358,
                1370506,
                1374024,
                1377911,
                1393848,
                1396334,
                1398874,
                1419371,
            ]
        )
        or (flag == "approval-comm-beta+" and obj_id == 8972248)
        or (flag == "testcase+" and obj_id == 267645)
        or (
            flag
            in [
                "platform-rel?",
                "blocking0.3-",
                "blocking-aviary1.0RC1-",
                "blocking-aviary1.1+",
                "blocking-firefox3.1-",
                "blocking1.8b4-",
                "blocking1.9+",
                "blocking1.9.0.3?",
                "blocking1.9.0.10+",
                "blocking1.9.0.17+",
                "wanted-firefox3.1?",
                "approval1.7.x+",
                "approval1.9.0.10+",
            ]
        )  # These flags have been removed.
    )


def rollback(bug, when=None, do_assert=False):
    def assert_or_log(msg):
        msg = f'{msg}, in bug {bug["id"]}'
        if do_assert:
            assert False, msg
        else:
            print(msg)

    def parse_flag_change(change):
        parts = change.split("(")
        if len(parts) != 1 and len(parts) != 2:
            assert_or_log(f"Too many parts for {change}")
            return None, None, None

        name_and_status = parts[0]
        name = name_and_status[:-1]
        status = name_and_status[-1]
        if status not in ["?", "+", "-"]:
            assert_or_log(f"unexpected status: {status}")
            return None, None, None

        requestee = None if len(parts) != 2 else parts[1][:-1]
        return name, status, requestee

    last_product = bug["product"]

    change_to_return = None
    if when is not None:
        for history in bug["history"]:
            for change in history["changes"]:
                if when(change):
                    change_to_return = change
                    rollback_date = dateutil.parser.parse(history["when"])
                    break

            if change_to_return is not None:
                break

        if change_to_return is None:
            return bug
    else:
        rollback_date = dateutil.parser.parse(bug["creation_time"])

    ret = False

    for history in reversed(bug["history"]):
        # TODO: Handle changes to product and component.
        # TODO: This code might be removed when https://bugzilla.mozilla.org/show_bug.cgi?id=1513952 is fixed.

        if ret:
            break

        for change in history["changes"]:
            if change is change_to_return:
                ret = True
                break

            field = change["field_name"]

            if field in "component":
                # TODO: Ignore this for now, not so easy to make it work https://bugzilla.mozilla.org/show_bug.cgi?id=1513952.
                continue

            if field == "qa_contact":
                # TODO: Ignore this for now. Example usage in 92144.
                continue

            if field == "cf_fx_iteration":
                # TODO: Ignore this for now. Example usage in 1101478.
                continue

            if field == "cf_crash_signature":
                # TODO: Ignore this for now. Example usage in 1437575.
                continue

            if field == "cf_backlog":
                # TODO: Ignore this for now. Example usage in 1048455.
                continue

            if field == "bug_mentor":
                # TODO: Ignore this for now. Example usage in 1042103.
                continue

            if field == "cf_user_story":
                # TODO: Ignore this for now. Example usage in 1369255.
                # Seems to be broken in Bugzilla.
                continue

            if field == "cf_rank":
                # TODO: Ignore this for now. Example usage in 1475099.
                continue

            if field in ["alias", "restrict_comments"]:
                continue

            if field == "longdescs.isprivate":
                # Ignore for now.
                continue

            if field == "version":
                # TODO: Ignore this for now. Example usage in 1162372 or 1389926.
                continue

            if "attachment_id" in change and field.startswith("attachments"):
                # TODO: Ignore changes to attachments for now.
                continue

            if field == "flagtypes.name":
                if "attachment_id" in change:
                    # https://bugzilla.mozilla.org/show_bug.cgi?id=1516172
                    if bug["id"] == 1_421_395:
                        continue

                    obj = None
                    for attachment in bug["attachments"]:
                        if attachment["id"] == change["attachment_id"]:
                            obj = attachment
                            break

                    if obj is None:
                        assert_or_log(f'Attachment {change["attachment_id"]} not found')
                        continue
                else:
                    obj = bug

                if change["added"]:
                    for to_remove in change["added"].split(", "):
                        # TODO: Skip needinfo/reviews for now, we need a way to match them precisely when there are multiple needinfos/reviews requested.
                        is_question_flag = any(
                            to_remove.startswith(s)
                            for s in [
                                "needinfo",
                                "review",
                                "feedback",
                                "ui-review",
                                "sec-approval",
                                "sec-review",
                                "data-review",
                                "approval-mozilla-",
                            ]
                        )

                        name, status, requestee = parse_flag_change(to_remove)

                        found_flag = None
                        for f in obj["flags"]:
                            if (
                                f["name"] == name
                                and f["status"] == status
                                and (
                                    requestee is None
                                    or (
                                        "requestee" in f and f["requestee"] == requestee
                                    )
                                )
                            ):
                                if (
                                    found_flag is not None
                                    and not is_expected_inconsistent_change_flag(
                                        to_remove, obj["id"]
                                    )
                                    and not is_question_flag
                                ):
                                    flag_text = "{}{}".format(f["name"], f["status"])
                                    if "requestee" in f:
                                        flag_text = "{}{}".format(
                                            flag_text, f["requestee"]
                                        )
                                    assert_or_log(f"{flag_text} found twice!")
                                found_flag = f

                        if found_flag is not None:
                            obj["flags"].remove(found_flag)
                        elif (
                            not is_expected_inconsistent_change_flag(
                                to_remove, obj["id"]
                            )
                            and not is_question_flag
                        ):
                            assert_or_log(
                                f"flag {to_remove} not found, in obj {obj['id']}"
                            )

                if change["removed"]:
                    # Inconsistent review and needinfo flags.
                    if bug["id"] in [785931, 1_342_178]:
                        continue

                    for to_add in change["removed"].split(", "):
                        name, status, requestee = parse_flag_change(to_add)

                        new_flag = {"name": name, "status": status}
                        if requestee is not None:
                            new_flag["requestee"] = requestee

                        obj["flags"].append(new_flag)

                continue

            # We don't support comment tags yet.
            if field == "comment_tag":
                continue

            if field == "comment_revision":
                obj = None
                for comment in bug["comments"]:
                    if comment["id"] == change["comment_id"]:
                        obj = comment
                        break

                if obj is None:
                    if change["comment_id"] != 14096735:
                        assert_or_log(f'Comment {change["comment_id"]} not found')
                    continue

                if obj["count"] != change["comment_count"]:
                    assert_or_log("Wrong comment count")

                # TODO: It should actually be applied on "raw_text".
                # if obj["text"] != change["added"]:
                #     assert_or_log(f"Current value for comment: ({obj['text']}) is different from previous value: ({change['added']}")

                obj["text"] = change["removed"]

                continue

            if change["added"] != "---":
                if field not in bug and not is_expected_inconsistent_field(
                    field, last_product, bug["id"]
                ):
                    assert_or_log(f"{field} is not present")

            if field in bug and isinstance(bug[field], list):
                if change["added"]:
                    for to_remove in change["added"].split(", "):
                        if field in FIELD_TYPES:
                            try:
                                to_remove = FIELD_TYPES[field](to_remove)
                            except Exception:
                                assert_or_log(
                                    f"Exception while transforming {to_remove} from {bug[field]} (field {field})"
                                )

                        if to_remove in bug[field]:
                            bug[field].remove(to_remove)
                        elif not is_expected_inconsistent_change_list_field(
                            field, bug["id"], to_remove
                        ):
                            assert_or_log(
                                f"{to_remove} is not in {bug[field]}, for field {field}"
                            )

                if change["removed"]:
                    for to_add in change["removed"].split(", "):
                        if field in FIELD_TYPES:
                            try:
                                to_add = FIELD_TYPES[field](to_add)
                            except Exception:
                                assert_or_log(
                                    f"Exception while transforming {to_add} from {bug[field]} (field {field})"
                                )
                        bug[field].append(to_add)
            else:
                if field in FIELD_TYPES:
                    try:
                        old_value = FIELD_TYPES[field](change["removed"])
                    except Exception:
                        assert_or_log(
                            f"Exception while transforming {change['removed']} from {bug[field]} (field {field})"
                        )
                    try:
                        new_value = FIELD_TYPES[field](change["added"])
                    except Exception:
                        assert_or_log(
                            f"Exception while transforming {change['added']} from {bug[field]} (field {field})"
                        )
                else:
                    old_value = change["removed"]
                    new_value = change["added"]

                if (
                    field in bug
                    and bug[field] != new_value
                    and not is_expected_inconsistent_change_field(
                        field, bug["id"], new_value, bug[field]
                    )
                ):
                    assert_or_log(
                        f"Current value for field {field}: ({bug[field]}) is different from previous value: ({new_value})"
                    )

                bug[field] = old_value

    if len(bug["comments"]) == 0:
        assert_or_log("There must be at least one comment")
        bug["comments"] = [
            {
                "count": 0,
                "id": 0,
                "text": "",
                "author": bug["creator"],
                "creation_time": bug["creation_time"],
            }
        ]

    # If the first comment is hidden.
    if bug["comments"][0]["count"] != 0:
        bug["comments"].insert(
            0,
            {
                "id": 0,
                "text": "",
                "author": bug["creator"],
                "creation_time": bug["creation_time"],
            },
        )

    bug["comments"] = [
        c
        for c in bug["comments"]
        if dateutil.parser.parse(c["creation_time"]) - relativedelta(seconds=3)
        <= rollback_date
    ]
    bug["attachments"] = [
        a
        for a in bug["attachments"]
        if dateutil.parser.parse(a["creation_time"]) - relativedelta(seconds=3)
        <= rollback_date
    ]

    return bug


def get_inconsistencies(bugs):
    inconsistencies = []

    for bug in bugs:
        try:
            rollback(bug, do_assert=True)
        except Exception as e:
            print(bug["id"])
            print(e)
            inconsistencies.append(bug)

    return inconsistencies


if __name__ == "__main__":
    import argparse

    from tqdm import tqdm

    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", help="Verbose mode", action="store_true")
    args = parser.parse_args()

    for bug in tqdm(bugzilla.get_bugs()):
        if args.verbose:
            print(bug["id"])

        rollback(bug, do_assert=True)



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from sklearn.ensemble import VotingClassifier
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import OrdinalEncoder

from bugbug import bug_features
from bugbug.models.component import ComponentModel
from bugbug.nn import KerasClassifier, KerasTextToSequences
from bugbug.utils import (
    DictExtractor,
    MissingOrdinalEncoder,
    StructuredColumnTransformer,
)

OPT_MSG_MISSING = (
    "Optional dependencies are missing, install them with: pip install bugbug[nn]\n"
)

try:
    from tensorflow.keras import Input, layers
    from tensorflow.keras.layers import (
        GRU,
        Bidirectional,
        Dense,
        Dropout,
        Embedding,
        Flatten,
        GlobalMaxPooling1D,
        SpatialDropout1D,
    )
    from tensorflow.keras.models import Model as KerasModel
except ImportError:
    raise ImportError(OPT_MSG_MISSING)


class ComponentNNClassifier(KerasClassifier):
    def __init__(self, **kwargs):
        # (epochs, batch_size) combinations
        fit_params = [(2, 256), (2, 512), (1, 1024)]
        super().__init__(fit_params=fit_params)

        self.set_params(**kwargs)

    def get_params(self, deep=True):
        return {
            "short_desc_maxlen": self.short_desc_maxlen,
            "short_desc_vocab_size": self.short_desc_vocab_size,
            "short_desc_emb_sz": self.short_desc_emb_sz,
            "long_desc_maxlen": self.long_desc_maxlen,
            "long_desc_vocab_size": self.long_desc_vocab_size,
            "long_desc_emb_sz": self.long_desc_emb_sz,
            "params": self.model_params,
        }

    def set_params(self, **kwargs):
        self.short_desc_maxlen = kwargs.pop("short_desc_maxlen")
        self.short_desc_vocab_size = kwargs.pop("short_desc_vocab_size")
        self.short_desc_emb_sz = kwargs.pop("short_desc_emb_sz")
        self.long_desc_maxlen = kwargs.pop("long_desc_maxlen")
        self.long_desc_vocab_size = kwargs.pop("long_desc_vocab_size")
        self.long_desc_emb_sz = kwargs.pop("long_desc_emb_sz")
        self.model_params = kwargs.pop("params")

        for (k, v) in self.model_params.items():
            setattr(self, k, v)

        return self

    def model_creator(self, X, y):
        short_desc_inp = Input(shape=(self.short_desc_maxlen,), name="title_sequence")
        short_desc_emb = Embedding(self.short_desc_vocab_size, self.short_desc_emb_sz)(
            short_desc_inp
        )
        short_desc_emb = SpatialDropout1D(self.short_desc_emb_dropout_rate)(
            short_desc_emb
        )
        short_desc_encoded = Bidirectional(
            GRU(
                self.short_desc_encoded_gru_units,
                dropout=self.short_desc_encoded_gru_dropout,
                recurrent_dropout=self.short_desc_encoded_recurrent_dropout,
                return_sequences=True,
            )
        )(short_desc_emb)
        short_desc_encoded = GlobalMaxPooling1D()(short_desc_encoded)

        long_desc_inp = Input(
            shape=(self.long_desc_maxlen,), name="first_comment_sequence"
        )
        long_desc_emb = Embedding(self.long_desc_vocab_size, self.long_desc_emb_sz)(
            long_desc_inp
        )
        long_desc_emb = SpatialDropout1D(self.long_desc_emb_dropout_rate)(long_desc_emb)
        long_desc_encoded = Bidirectional(
            GRU(
                self.long_desc_encoded_gru_units,
                dropout=self.long_desc_encoded_dropout,
                recurrent_dropout=self.long_desc_encoded_recurrent_dropout,
                return_sequences=True,
            )
        )(long_desc_emb)
        long_desc_encoded = GlobalMaxPooling1D()(long_desc_encoded)

        rep_platform_inp = Input(shape=(1,), name="platform")
        rep_platform_emb = Embedding(
            input_dim=self.rep_platform_emb_input_dim,
            output_dim=self.rep_platform_emb_output_dim,
            input_length=1,
        )(rep_platform_inp)
        rep_platform_emb = SpatialDropout1D(self.rep_platform_emb_spatial_dropout_rate)(
            rep_platform_emb
        )
        rep_platform_emb = Flatten()(rep_platform_emb)
        rep_platform_emb = Dropout(self.rep_platform_emb_dropout_rate)(rep_platform_emb)

        op_sys_inp = Input(shape=(1,), name="op_sys")
        op_sys_emb = Embedding(
            input_dim=self.op_sys_emb_input_dim,
            output_dim=self.op_sys_emb_output_dim,
            input_length=1,
        )(op_sys_inp)
        op_sys_emb = SpatialDropout1D(self.op_sys_emb_spatial_dropout_rate)(op_sys_emb)
        op_sys_emb = Flatten()(op_sys_emb)
        op_sys_emb = Dropout(self.op_sys_emb_dropout_rate)(op_sys_emb)

        reporter_inp = Input(shape=(1,), name="bug_reporter")
        reporter_emb = Embedding(
            input_dim=self.reporter_emb_input_dim,
            output_dim=self.reporter_emb_output_dim,
            input_length=1,
        )(reporter_inp)
        reporter_emb = SpatialDropout1D(self.reporter_emb_spatial_dropout_rate)(
            reporter_emb
        )
        reporter_emb = Flatten()(reporter_emb)
        reporter_emb = Dropout(self.reporter_emb_dropout_rate)(reporter_emb)

        tfidf_word_inp = Input(
            shape=(X["title_word_tfidf"].shape[1],), name="title_word_tfidf"
        )
        tfidf_word = Dense(self.tfidf_word_dense_units, activation="relu")(
            tfidf_word_inp
        )
        tfidf_word = Dropout(self.tfidf_word_dropout_rate)(tfidf_word)

        tfidf_char_inp = Input(
            shape=(X["title_char_tfidf"].shape[1],), name="title_char_tfidf"
        )
        tfidf_char = Dense(self.tfidf_char_inp_dense_unit, activation="relu")(
            tfidf_char_inp
        )
        tfidf_char = Dropout(self.tfidf_char_inp_dropout_rate)(tfidf_char)

        x = layers.concatenate(
            [
                short_desc_encoded,
                long_desc_encoded,
                rep_platform_emb,
                op_sys_emb,
                reporter_emb,
                tfidf_word,
                tfidf_char,
            ],
            axis=-1,
        )

        x = Dense(self.x_dense_unit, activation="relu")(x)
        x = Dropout(self.x_dropout_rate)(x)
        x = Dense(y.shape[1], activation="softmax")(x)

        model = KerasModel(
            [
                short_desc_inp,
                long_desc_inp,
                rep_platform_inp,
                op_sys_inp,
                reporter_inp,
                tfidf_word_inp,
                tfidf_char_inp,
            ],
            x,
        )
        model.compile(
            optimizer="adam", loss=["categorical_crossentropy"], metrics=["acc"]
        )

        return model


class ComponentNNModel(ComponentModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.short_desc_maxlen = 20
        self.short_desc_vocab_size = 25000
        self.short_desc_emb_sz = 300
        self.long_desc_maxlen = 100
        self.long_desc_vocab_size = 25000
        self.long_desc_emb_sz = 300
        self.cross_validation_enabled = False

        self.params = [
            {
                "short_desc_emb_dropout_rate": 0.2,
                "short_desc_encoded_gru_units": 256,
                "short_desc_encoded_gru_dropout": 0.45,
                "short_desc_encoded_recurrent_dropout": 0.5,
                "long_desc_emb_dropout_rate": 0.25,
                "long_desc_encoded_gru_units": 256,
                "long_desc_encoded_dropout": 0.5,
                "long_desc_encoded_recurrent_dropout": 0.55,
                "rep_platform_emb_input_dim": 14,
                "rep_platform_emb_output_dim": 25,
                "rep_platform_emb_spatial_dropout_rate": 0.1,
                "rep_platform_emb_dropout_rate": 0.45,
                "op_sys_emb_input_dim": 48,
                "op_sys_emb_output_dim": 50,
                "op_sys_emb_spatial_dropout_rate": 0.1,
                "op_sys_emb_dropout_rate": 0.45,
                "reporter_emb_input_dim": 46544,
                "reporter_emb_output_dim": 100,
                "reporter_emb_spatial_dropout_rate": 0.15,
                "reporter_emb_dropout_rate": 0.5,
                "tfidf_word_dense_units": 600,
                "tfidf_word_dropout_rate": 0.5,
                "tfidf_char_inp_dense_unit": 500,
                "tfidf_char_inp_dropout_rate": 0.5,
                "x_dense_unit": 2000,
                "x_dropout_rate": 0.6,
            },
            {
                "short_desc_emb_dropout_rate": 0.2,
                "short_desc_encoded_gru_units": 250,
                "short_desc_encoded_gru_dropout": 0.45,
                "short_desc_encoded_recurrent_dropout": 0.45,
                "long_desc_emb_dropout_rate": 0.25,
                "long_desc_encoded_gru_units": 250,
                "long_desc_encoded_dropout": 0.45,
                "long_desc_encoded_recurrent_dropout": 0.45,
                "rep_platform_emb_input_dim": 14,
                "rep_platform_emb_output_dim": 30,
                "rep_platform_emb_spatial_dropout_rate": 0.1,
                "rep_platform_emb_dropout_rate": 0.4,
                "op_sys_emb_input_dim": 48,
                "op_sys_emb_output_dim": 55,
                "op_sys_emb_spatial_dropout_rate": 0.1,
                "op_sys_emb_dropout_rate": 0.4,
                "reporter_emb_input_dim": 46544,
                "reporter_emb_output_dim": 110,
                "reporter_emb_spatial_dropout_rate": 0.15,
                "reporter_emb_dropout_rate": 0.45,
                "tfidf_word_dense_units": 610,
                "tfidf_word_dropout_rate": 0.45,
                "tfidf_char_inp_dense_unit": 510,
                "tfidf_char_inp_dropout_rate": 0.5,
                "x_dense_unit": 1970,
                "x_dropout_rate": 0.5,
            },
            {
                "short_desc_emb_dropout_rate": 0.2,
                "short_desc_encoded_gru_units": 266,
                "short_desc_encoded_gru_dropout": 0.45,
                "short_desc_encoded_recurrent_dropout": 0.45,
                "long_desc_emb_dropout_rate": 0.25,
                "long_desc_encoded_gru_units": 266,
                "long_desc_encoded_dropout": 0.45,
                "long_desc_encoded_recurrent_dropout": 0.55,
                "rep_platform_emb_input_dim": 14,
                "rep_platform_emb_output_dim": 35,
                "rep_platform_emb_spatial_dropout_rate": 0.1,
                "rep_platform_emb_dropout_rate": 0.45,
                "op_sys_emb_input_dim": 48,
                "op_sys_emb_output_dim": 60,
                "op_sys_emb_spatial_dropout_rate": 0.1,
                "op_sys_emb_dropout_rate": 0.45,
                "reporter_emb_input_dim": 46544,
                "reporter_emb_output_dim": 120,
                "reporter_emb_spatial_dropout_rate": 0.15,
                "reporter_emb_dropout_rate": 0.45,
                "tfidf_word_dense_units": 620,
                "tfidf_word_dropout_rate": 0.5,
                "tfidf_char_inp_dense_unit": 520,
                "tfidf_char_inp_dropout_rate": 0.45,
                "x_dense_unit": 1950,
                "x_dropout_rate": 0.5,
            },
        ]

        feature_extractors = [
            bug_features.bug_reporter(),
            bug_features.platform(),
            bug_features.op_sys(),
        ]

        cleanup_functions = []

        self.extraction_pipeline = Pipeline(
            [
                (
                    "bug_extractor",
                    bug_features.BugExtractor(feature_extractors, cleanup_functions),
                ),
                (
                    "union",
                    StructuredColumnTransformer(
                        [
                            (
                                "platform",
                                make_pipeline(
                                    DictExtractor("platform"), OrdinalEncoder()
                                ),
                                "data",
                            ),
                            (
                                "op_sys",
                                make_pipeline(
                                    DictExtractor("op_sys"), OrdinalEncoder()
                                ),
                                "data",
                            ),
                            (
                                "bug_reporter",
                                make_pipeline(
                                    DictExtractor("bug_reporter"),
                                    MissingOrdinalEncoder(),
                                ),
                                "data",
                            ),
                            (
                                "title_sequence",
                                KerasTextToSequences(
                                    self.short_desc_maxlen, self.short_desc_vocab_size
                                ),
                                "title",
                            ),
                            (
                                "first_comment_sequence",
                                KerasTextToSequences(
                                    self.long_desc_maxlen, self.long_desc_vocab_size
                                ),
                                "first_comment",
                            ),
                            (
                                "title_char_tfidf",
                                TfidfVectorizer(
                                    strip_accents="unicode",
                                    analyzer="char",
                                    stop_words="english",
                                    ngram_range=(2, 4),
                                    max_features=25000,
                                    sublinear_tf=True,
                                ),
                                "title",
                            ),
                            (
                                "title_word_tfidf",
                                TfidfVectorizer(
                                    strip_accents="unicode",
                                    min_df=0.0001,
                                    max_df=0.1,
                                    analyzer="word",
                                    token_pattern=r"\w{1,}",
                                    stop_words="english",
                                    ngram_range=(2, 4),
                                    max_features=30000,
                                    sublinear_tf=True,
                                ),
                                "title",
                            ),
                        ]
                    ),
                ),
            ]
        )

        kwargs = {
            "short_desc_maxlen": self.short_desc_maxlen,
            "short_desc_vocab_size": self.short_desc_vocab_size,
            "short_desc_emb_sz": self.short_desc_emb_sz,
            "long_desc_maxlen": self.long_desc_maxlen,
            "long_desc_vocab_size": self.long_desc_vocab_size,
            "long_desc_emb_sz": self.long_desc_emb_sz,
        }

        estimators = []
        for i, params in enumerate(self.params):
            kwargs["params"] = params
            estimator = ComponentNNClassifier(**kwargs)
            estimators.append(("model_{}".format(i), estimator))

        self.clf = VotingClassifier(
            estimators=estimators, voting="soft", weights=[1, 1, 1]
        )

    def get_feature_names(self):
        return self.extraction_pipeline.named_steps["union"].named_transformers_.keys()



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from collections import defaultdict
from typing import Any, Dict, List, Tuple

import joblib
import matplotlib
import numpy as np
import shap
from imblearn.metrics import (
    classification_report_imbalanced,
    geometric_mean_score,
    make_index_balanced_accuracy,
    specificity_score,
)
from imblearn.pipeline import make_pipeline
from sklearn import metrics
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import precision_recall_fscore_support
from sklearn.model_selection import cross_validate, train_test_split
from tabulate import tabulate

from bugbug import bugzilla, db, repository
from bugbug.nlp import SpacyVectorizer
from bugbug.utils import split_tuple_generator, to_array


def classification_report_imbalanced_values(
    y_true, y_pred, labels, target_names=None, sample_weight=None, digits=2, alpha=0.1
):
    """Copy of imblearn.metrics.classification_report_imbalanced to have
    access to the raw values. The code is mostly the same except the
    formatting code and generation of the report which haven removed. Copied
    from version 0.4.3. The original code is living here:
    https://github.com/scikit-learn-contrib/imbalanced-learn/blob/b861b3a8e3414c52f40a953f2e0feca5b32e7460/imblearn/metrics/_classification.py#L790
    """
    labels = np.asarray(labels)

    if target_names is None:
        target_names = [str(label) for label in labels]

    # Compute the different metrics
    # Precision/recall/f1
    precision, recall, f1, support = precision_recall_fscore_support(
        y_true, y_pred, labels=labels, average=None, sample_weight=sample_weight
    )
    # Specificity
    specificity = specificity_score(
        y_true, y_pred, labels=labels, average=None, sample_weight=sample_weight
    )
    # Geometric mean
    geo_mean = geometric_mean_score(
        y_true, y_pred, labels=labels, average=None, sample_weight=sample_weight
    )
    # Index balanced accuracy
    iba_gmean = make_index_balanced_accuracy(alpha=alpha, squared=True)(
        geometric_mean_score
    )
    iba = iba_gmean(
        y_true, y_pred, labels=labels, average=None, sample_weight=sample_weight
    )

    result = {"targets": {}}

    for i, label in enumerate(labels):
        result["targets"][target_names[i]] = {
            "precision": precision[i],
            "recall": recall[i],
            "specificity": specificity[i],
            "f1": f1[i],
            "geo_mean": geo_mean[i],
            "iba": iba[i],
            "support": support[i],
        }

    result["average"] = {
        "precision": np.average(precision, weights=support),
        "recall": np.average(recall, weights=support),
        "specificity": np.average(specificity, weights=support),
        "f1": np.average(f1, weights=support),
        "geo_mean": np.average(geo_mean, weights=support),
        "iba": np.average(iba, weights=support),
        "support": np.sum(support),
    }

    return result


def print_labeled_confusion_matrix(confusion_matrix, labels, is_multilabel=False):
    confusion_matrix_table = confusion_matrix.tolist()

    # Don't show the Not classified row in the table output
    if "__NOT_CLASSIFIED__" in labels and not is_multilabel:
        confusion_matrix_table.pop(labels.index("__NOT_CLASSIFIED__"))

    if not is_multilabel:
        confusion_matrix_table = [confusion_matrix_table]

    for num, table in enumerate(confusion_matrix_table):
        if is_multilabel:
            print(f"label: {labels[num]}")
            table_labels = [0, 1]
        else:
            table_labels = labels

        confusion_matrix_header = []
        for i in range(len(table[0])):
            confusion_matrix_header.append(
                f"{table_labels[i]} (Predicted)"
                if table_labels[i] != "__NOT_CLASSIFIED__"
                else "Not classified"
            )
        for i in range(len(table)):
            table[i].insert(0, f"{table_labels[i]} (Actual)")
        print(
            tabulate(table, headers=confusion_matrix_header, tablefmt="fancy_grid"),
            end="\n\n",
        )


def sort_class_names(class_names):
    if len(class_names) == 2:
        class_names = sorted(list(class_names), reverse=True)
    else:
        class_names = sorted(list(class_names))

    return class_names


class Model:
    def __init__(self, lemmatization=False):
        if lemmatization:
            self.text_vectorizer = SpacyVectorizer
        else:
            self.text_vectorizer = TfidfVectorizer

        self.cross_validation_enabled = True
        self.sampler = None

        self.calculate_importance = True

        self.store_dataset = False

        self.entire_dataset_training = False

        # DBs required for training.
        self.training_dbs: List[str] = []
        # DBs and DB support files required at runtime.
        self.eval_dbs: Dict[str, Tuple[str, ...]] = {}

    def download_eval_dbs(
        self, extract: bool = True, ensure_exist: bool = True
    ) -> None:
        for eval_db, eval_files in self.eval_dbs.items():
            for eval_file in eval_files:
                if db.is_registered(eval_file):
                    assert db.download(eval_file, extract=extract) or not ensure_exist
                else:
                    assert (
                        db.download_support_file(eval_db, eval_file, extract=extract)
                        or not ensure_exist
                    )

    @property
    def le(self):
        """Classifier agnostic getter for the label encoder property"""
        try:
            return self.clf._le
        except AttributeError:
            return self.clf.le_

    def get_feature_names(self):
        return []

    def get_human_readable_feature_names(self):
        feature_names = self.get_feature_names()

        cleaned_feature_names = []
        for full_feature_name in feature_names:
            type_, feature_name = full_feature_name.split("__", 1)

            if type_ == "desc":
                feature_name = f"Description contains '{feature_name}'"
            elif type_ == "title":
                feature_name = f"Title contains '{feature_name}'"
            elif type_ == "first_comment":
                feature_name = f"First comment contains '{feature_name}'"
            elif type_ == "comments":
                feature_name = f"Comments contain '{feature_name}'"
            elif type_ == "text":
                feature_name = f"Combined text contains '{feature_name}'"
            elif type_ == "data":
                if " in " in feature_name and feature_name.endswith("=True"):
                    feature_name = feature_name[: -len("=True")]
            else:
                raise Exception(f"Unexpected feature type for: {full_feature_name}")

            cleaned_feature_names.append(feature_name)

        return cleaned_feature_names

    def get_important_features(self, cutoff, shap_values):
        # returns top features for a shap_value matrix
        def get_top_features(cutoff, shap_values):
            # Calculate the values that represent the fraction of the model output variability attributable
            # to each feature across the whole dataset.
            shap_sums = shap_values.sum(0)
            abs_shap_sums = np.abs(shap_values).sum(0)
            rel_shap_sums = abs_shap_sums / abs_shap_sums.sum()

            cut_off_value = cutoff * np.amax(rel_shap_sums)

            # Get indices of features that pass the cut off value
            top_feature_indices = np.where(rel_shap_sums >= cut_off_value)[0]
            # Get the importance values of the top features from their indices
            top_features = np.take(rel_shap_sums, top_feature_indices)
            # Gets the sign of the importance from shap_sums as boolean
            is_positive = (np.take(shap_sums, top_feature_indices)) >= 0
            # Stack the importance, indices and shap_sums in a 2D array
            top_features = np.column_stack(
                (top_features, top_feature_indices, is_positive)
            )
            # Sort the array (in decreasing order of importance values)
            top_features = top_features[top_features[:, 0].argsort()][::-1]

            return top_features

        important_features = {}
        important_features["classes"] = {}
        important_features["average"] = get_top_features(
            cutoff, np.sum(np.abs(shap_values), axis=0)
        )
        for num, item in enumerate(shap_values):
            # top features for that class
            top_item_features = get_top_features(cutoff, item)

            # shap values of top average features for that class
            abs_sums = np.abs(item).sum(0)
            rel_sums = abs_sums / abs_sums.sum()
            is_pos = ["+" if shap_sum >= 0 else "-" for shap_sum in item.sum(0)]
            top_avg = [
                is_pos[int(index)] + str(rel_sums[int(index)])
                for importance, index, is_positive in important_features["average"]
            ]

            class_name = self.le.inverse_transform([num])[0]

            important_features["classes"][class_name] = (
                top_item_features,
                top_avg,
            )

        return important_features

    def print_feature_importances(self, important_features, class_probabilities=None):
        feature_names = self.get_human_readable_feature_names()
        # extract importance values from the top features for the predicted class
        # when classifying
        if class_probabilities is not None:
            predicted_class_index = class_probabilities.argmax(axis=-1)[0]
            predicted_class = self.le.inverse_transform([predicted_class_index])[0]

            imp_values = important_features["classes"][predicted_class][0]
            shap_val = []
            top_feature_names = []
            for importance, index, is_positive in imp_values:
                if is_positive:
                    shap_val.append("+" + str(importance))
                else:
                    shap_val.append("-" + str(importance))

                feature_value = np.squeeze(
                    to_array(important_features["values"])[:, int(index)]
                )
                top_feature_names.append(
                    f"{feature_names[int(index)]} = {feature_value.round(decimals=5)}"
                )
            shap_val = [[predicted_class] + shap_val]

        # extract importance values from the top features for all the classes
        # when training
        else:
            top_feature_names = [
                feature_names[int(index)]
                for importance, index, is_pos in important_features["average"]
            ]
            shap_val = [
                [class_name] + imp_values[1]
                for class_name, imp_values in important_features["classes"].items()
            ]

        # allow maximum of 3 columns in a row to fit the page better
        COLUMNS = 3
        print("Top {} features:".format(len(top_feature_names)))
        for i in range(0, len(top_feature_names), COLUMNS):
            table = []
            for item in shap_val:
                table.append(item[i : i + COLUMNS])
            print(
                tabulate(
                    table,
                    headers=(["classes"] + top_feature_names)[i : i + COLUMNS],
                    tablefmt="grid",
                ),
                end="\n\n",
            )

    def save_feature_importances(self, important_features, feature_names):
        # Returns a JSON-encodable dictionary that can be saved in the metrics
        # report
        feature_report = {"classes": {}, "average": {}}
        top_feature_names = []

        for importance, index, is_pos in important_features["average"]:
            feature_name = feature_names[int(index)]

            top_feature_names.append(feature_name)
            feature_report["average"][feature_name] = importance

        for i, feature_name in enumerate(top_feature_names):
            for class_name, imp_values in important_features["classes"].items():
                class_report = feature_report["classes"].setdefault(
                    class_name.item(), {}
                )
                class_report[feature_name] = float(imp_values[1][i])

        return feature_report

    def train_test_split(self, X, y):
        return train_test_split(X, y, test_size=0.1, random_state=0)

    def evaluation(self):
        """Subclasses can implement their own additional evaluation."""
        pass

    def get_labels(self) -> Tuple[Dict[Any, Any], List[Any]]:
        """Subclasses implement their own function to gather labels."""
        pass

    def train(self, importance_cutoff=0.15, limit=None):
        classes, self.class_names = self.get_labels()
        self.class_names = sort_class_names(self.class_names)

        # Get items and labels, filtering out those for which we have no labels.
        X_gen, y = split_tuple_generator(lambda: self.items_gen(classes))

        # Extract features from the items.
        X = self.extraction_pipeline.fit_transform(X_gen)

        # Calculate labels.
        y = np.array(y)

        if limit:
            X = X[:limit]
            y = y[:limit]

        print(f"X: {X.shape}, y: {y.shape}")

        is_multilabel = isinstance(y[0], np.ndarray)
        is_binary = len(self.class_names) == 2

        # Split dataset in training and test.
        X_train, X_test, y_train, y_test = self.train_test_split(X, y)
        if self.sampler is not None:
            pipeline = make_pipeline(self.sampler, self.clf)
        else:
            pipeline = self.clf

        tracking_metrics = {}

        # Use k-fold cross validation to evaluate results.
        if self.cross_validation_enabled:
            scorings = ["accuracy"]
            if len(self.class_names) == 2:
                scorings += ["precision", "recall"]

            scores = cross_validate(pipeline, X_train, y_train, scoring=scorings, cv=5)

            print("Cross Validation scores:")
            for scoring in scorings:
                score = scores[f"test_{scoring}"]
                tracking_metrics[f"test_{scoring}"] = {
                    "mean": score.mean(),
                    "std": score.std() * 2,
                }
                print(
                    f"{scoring.capitalize()}: f{score.mean()} (+/- {score.std() * 2})"
                )

        print(f"X_train: {X_train.shape}, y_train: {y_train.shape}")

        # Training on the resampled dataset if sampler is provided.
        if self.sampler is not None:
            X_train, y_train = self.sampler.fit_resample(X_train, y_train)

            print(f"resampled X_train: {X_train.shape}, y_train: {y_train.shape}")

        print(f"X_test: {X_test.shape}, y_test: {y_test.shape}")

        self.clf.fit(X_train, y_train)

        print("Model trained")

        feature_names = self.get_human_readable_feature_names()
        if self.calculate_importance and len(feature_names):
            explainer = shap.TreeExplainer(self.clf)
            shap_values = explainer.shap_values(X_train)

            # In the binary case, sometimes shap returns a single shap values matrix.
            if is_binary and not isinstance(shap_values, list):
                shap_values = [-shap_values, shap_values]
                summary_plot_value = shap_values[1]
                summary_plot_type = "layered_violin"
            else:
                summary_plot_value = shap_values
                summary_plot_type = None

            shap.summary_plot(
                summary_plot_value,
                to_array(X_train),
                feature_names=feature_names,
                class_names=self.class_names,
                plot_type=summary_plot_type,
                show=False,
            )

            matplotlib.pyplot.savefig("feature_importance.png", bbox_inches="tight")
            matplotlib.pyplot.xlabel("Impact on model output")
            matplotlib.pyplot.clf()

            important_features = self.get_important_features(
                importance_cutoff, shap_values
            )

            self.print_feature_importances(important_features)

            # Save the important features in the metric report too
            feature_report = self.save_feature_importances(
                important_features, feature_names
            )

            tracking_metrics["feature_report"] = feature_report

        print("Training Set scores:")
        y_pred = self.clf.predict(X_train)
        if not is_multilabel:
            print(
                classification_report_imbalanced(
                    y_train, y_pred, labels=self.class_names
                )
            )

        print("Test Set scores:")
        # Evaluate results on the test set.
        y_pred = self.clf.predict(X_test)

        if is_multilabel:
            assert isinstance(
                y_pred[0], np.ndarray
            ), "The predictions should be multilabel"

        print(f"No confidence threshold - {len(y_test)} classified")
        if is_multilabel:
            confusion_matrix = metrics.multilabel_confusion_matrix(y_test, y_pred)
        else:
            confusion_matrix = metrics.confusion_matrix(
                y_test, y_pred, labels=self.class_names
            )

            print(
                classification_report_imbalanced(
                    y_test, y_pred, labels=self.class_names
                )
            )
            report = classification_report_imbalanced_values(
                y_test, y_pred, labels=self.class_names
            )

            tracking_metrics["report"] = report

        print_labeled_confusion_matrix(
            confusion_matrix, self.class_names, is_multilabel=is_multilabel
        )

        tracking_metrics["confusion_matrix"] = confusion_matrix.tolist()

        confidence_thresholds = [0.6, 0.7, 0.8, 0.9]

        if is_binary:
            confidence_thresholds = [0.1, 0.2, 0.3, 0.4] + confidence_thresholds

        # Evaluate results on the test set for some confidence thresholds.
        for confidence_threshold in confidence_thresholds:
            y_pred_probas = self.clf.predict_proba(X_test)
            confidence_class_names = self.class_names + ["__NOT_CLASSIFIED__"]

            y_pred_filter = []
            classified_indices = []
            for i in range(0, len(y_test)):
                if not is_binary:
                    argmax = np.argmax(y_pred_probas[i])
                else:
                    argmax = 1 if y_pred_probas[i][1] > confidence_threshold else 0

                if y_pred_probas[i][argmax] < confidence_threshold:
                    if not is_multilabel:
                        y_pred_filter.append("__NOT_CLASSIFIED__")
                    continue

                classified_indices.append(i)
                if is_multilabel:
                    y_pred_filter.append(y_pred[i])
                else:
                    y_pred_filter.append(argmax)

            if not is_multilabel:
                y_pred_filter = np.array(y_pred_filter)
                y_pred_filter[classified_indices] = self.le.inverse_transform(
                    np.array(y_pred_filter[classified_indices], dtype=int)
                )

            classified_num = sum(1 for v in y_pred_filter if v != "__NOT_CLASSIFIED__")

            print(
                f"\nConfidence threshold > {confidence_threshold} - {classified_num} classified"
            )
            if is_multilabel:
                confusion_matrix = metrics.multilabel_confusion_matrix(
                    y_test[classified_indices], np.asarray(y_pred_filter)
                )
            else:
                confusion_matrix = metrics.confusion_matrix(
                    y_test.astype(str),
                    y_pred_filter.astype(str),
                    labels=confidence_class_names,
                )
                print(
                    classification_report_imbalanced(
                        y_test.astype(str),
                        y_pred_filter.astype(str),
                        labels=confidence_class_names,
                    )
                )
            print_labeled_confusion_matrix(
                confusion_matrix, confidence_class_names, is_multilabel=is_multilabel
            )

        self.evaluation()

        if self.entire_dataset_training:
            print("Retraining on the entire dataset...")

            if self.sampler is not None:
                X_train, y_train = self.sampler.fit_resample(X, y)
            else:
                X_train = X
                y_train = y

            print(f"X_train: {X_train.shape}, y_train: {y_train.shape}")

            self.clf.fit(X_train, y_train)

        joblib.dump(self, self.__class__.__name__.lower())
        if self.store_dataset:
            joblib.dump(X, f"{self.__class__.__name__.lower()}_data_X")
            joblib.dump(y, f"{self.__class__.__name__.lower()}_data_y")

        return tracking_metrics

    @staticmethod
    def load(model_file_name):
        return joblib.load(model_file_name)

    def overwrite_classes(self, items, classes, probabilities):
        return classes

    def classify(
        self,
        items,
        probabilities=False,
        importances=False,
        importance_cutoff=0.15,
        background_dataset=None,
    ):
        assert items is not None
        assert (
            self.extraction_pipeline is not None and self.clf is not None
        ), "The module needs to be initialized first"

        if not isinstance(items, list):
            items = [items]

        assert isinstance(items[0], dict) or isinstance(items[0], tuple)

        X = self.extraction_pipeline.transform(lambda: items)
        if probabilities:
            classes = self.clf.predict_proba(X)
        else:
            classes = self.clf.predict(X)

        classes = self.overwrite_classes(items, classes, probabilities)

        if importances:
            pred_class_index = classes.argmax(axis=-1)[0]
            pred_class = self.le.inverse_transform([pred_class_index])[0]

            if background_dataset is None:
                explainer = shap.TreeExplainer(self.clf)
            else:
                explainer = shap.TreeExplainer(
                    self.clf,
                    to_array(background_dataset(pred_class)),
                    feature_perturbation="interventional",
                )

            shap_values = explainer.shap_values(to_array(X))

            # In the binary case, sometimes shap returns a single shap values matrix.
            if len(classes[0]) == 2 and not isinstance(shap_values, list):
                shap_values = [-shap_values, shap_values]

            important_features = self.get_important_features(
                importance_cutoff, shap_values
            )
            important_features["values"] = X

            top_indexes = [
                int(index)
                for _, index, _ in important_features["classes"][pred_class][0]
            ]

            feature_names = self.get_human_readable_feature_names()

            feature_legend = {
                str(i + 1): feature_names[feature_i]
                for i, feature_i in enumerate(top_indexes)
            }

            return (
                classes,
                {"importances": important_features, "feature_legend": feature_legend},
            )

        return classes

    def check(self):
        """Subclasses can implement their own check, the base model doesn't
        check anything at the moment
        """
        return True

    def get_extra_data(self):
        """Returns a dict that can be used for customers who need static
        extra data for a given model. Must return a dict and JSON-encodable
        types.
        """
        return {}


class BugModel(Model):
    def __init__(self, lemmatization=False, commit_data=False):
        Model.__init__(self, lemmatization)
        self.commit_data = commit_data
        self.training_dbs = [bugzilla.BUGS_DB]
        if commit_data:
            self.training_dbs.append(repository.COMMITS_DB)

    def items_gen(self, classes):
        if not self.commit_data:
            commit_map = None
        else:
            commit_map = defaultdict(list)

            for commit in repository.get_commits():
                bug_id = commit["bug_id"]
                if not bug_id:
                    continue

                commit_map[bug_id].append(commit)

            assert len(commit_map) > 0

        for bug in bugzilla.get_bugs():
            bug_id = bug["id"]
            if bug_id not in classes:
                continue

            if self.commit_data:
                if bug_id in commit_map:
                    bug["commits"] = commit_map[bug_id]
                else:
                    bug["commits"] = []

            yield bug, classes[bug_id]


class CommitModel(Model):
    def __init__(self, lemmatization=False, bug_data=False):
        Model.__init__(self, lemmatization)
        self.bug_data = bug_data
        self.training_dbs = [repository.COMMITS_DB]
        if bug_data:
            self.training_dbs.append(bugzilla.BUGS_DB)

    def items_gen(self, classes):
        if not self.bug_data:
            bug_map = None
        else:
            all_bug_ids = set(
                commit["bug_id"]
                for commit in repository.get_commits()
                if commit["node"] in classes
            )

            bug_map = {
                bug["id"]: bug
                for bug in bugzilla.get_bugs()
                if bug["id"] in all_bug_ids
            }

            assert len(bug_map) > 0

        for commit in repository.get_commits():
            if commit["node"] not in classes:
                continue

            if self.bug_data:
                if commit["bug_id"] in bug_map:
                    commit["bug"] = bug_map[commit["bug_id"]]
                else:
                    commit["bug"] = {}

            yield commit, classes[commit["node"]]


class BugCoupleModel(Model):
    def __init__(self, lemmatization=False):
        Model.__init__(self, lemmatization)
        self.training_dbs = [bugzilla.BUGS_DB]

    def items_gen(self, classes):
        bugs = {}
        for bug in bugzilla.get_bugs():
            bugs[bug["id"]] = bug

        for (bug_id1, bug_id2), label in classes.items():
            yield (bugs[bug_id1], bugs[bug_id2]), label



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import collections
import concurrent.futures
import logging
import math
import pickle
import statistics
from functools import reduce
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
)

import numpy as np
import xgboost
from imblearn.under_sampling import RandomUnderSampler
from ortools.linear_solver import pywraplp
from sklearn.compose import ColumnTransformer
from sklearn.feature_extraction import DictVectorizer
from sklearn.pipeline import Pipeline
from tqdm import tqdm

from bugbug import (
    commit_features,
    repository,
    test_scheduling,
    test_scheduling_features,
    utils,
)
from bugbug.model import Model

logger = logging.getLogger(__name__)


def get_commit_map(
    revs: Optional[Set[test_scheduling.Revision]] = None,
) -> Dict[test_scheduling.Revision, repository.CommitDict]:
    commit_map = {}

    for commit in repository.get_commits():
        if revs is not None and commit["node"] not in revs:
            continue

        commit_map[commit["node"]] = commit

    assert len(commit_map) > 0
    return commit_map


class TestSelectModel(Model):
    def __init__(self, lemmatization=False, granularity="label", failures_skip=None):
        Model.__init__(self, lemmatization)

        self.granularity = granularity
        self.failures_skip = failures_skip

        self.training_dbs = [repository.COMMITS_DB]
        self.eval_dbs[repository.COMMITS_DB] = (
            repository.COMMITS_DB,
            repository.COMMIT_EXPERIENCES_DB,
        )
        if granularity == "label":
            self.training_dbs.append(test_scheduling.TEST_LABEL_SCHEDULING_DB)
            self.eval_dbs[test_scheduling.TEST_LABEL_SCHEDULING_DB] = (
                test_scheduling.PAST_FAILURES_LABEL_DB,
                test_scheduling.FAILING_TOGETHER_LABEL_DB,
            )
        elif granularity == "group":
            self.training_dbs.append(test_scheduling.TEST_GROUP_SCHEDULING_DB)
            self.eval_dbs[test_scheduling.TEST_GROUP_SCHEDULING_DB] = (
                test_scheduling.PAST_FAILURES_GROUP_DB,
                test_scheduling.TOUCHED_TOGETHER_DB,
            )
            self.eval_dbs[test_scheduling.TEST_CONFIG_GROUP_SCHEDULING_DB] = (
                test_scheduling.FAILING_TOGETHER_CONFIG_GROUP_DB,
            )
        elif granularity == "config_group":
            self.training_dbs.append(test_scheduling.TEST_CONFIG_GROUP_SCHEDULING_DB)
            self.eval_dbs[test_scheduling.TEST_CONFIG_GROUP_SCHEDULING_DB] = (
                test_scheduling.PAST_FAILURES_CONFIG_GROUP_DB,
                test_scheduling.TOUCHED_TOGETHER_DB,
            )

        self.cross_validation_enabled = False

        self.entire_dataset_training = True

        self.sampler = RandomUnderSampler(random_state=0)

        feature_extractors = [
            test_scheduling_features.prev_failures(),
        ]

        if granularity == "label":
            feature_extractors += [
                test_scheduling_features.platform(),
                # test_scheduling_features.chunk(),
                test_scheduling_features.suite(),
            ]
        elif granularity in ("group", "config_group"):
            feature_extractors += [
                test_scheduling_features.path_distance(),
                test_scheduling_features.common_path_components(),
                test_scheduling_features.touched_together(),
            ]

        self.extraction_pipeline = Pipeline(
            [
                (
                    "commit_extractor",
                    commit_features.CommitExtractor(feature_extractors, []),
                ),
                ("union", ColumnTransformer([("data", DictVectorizer(), "data")])),
            ]
        )

        self.clf = xgboost.XGBClassifier(n_jobs=utils.get_physical_cpu_count())
        self.clf.set_params(predictor="cpu_predictor")

    def get_pushes(
        self, apply_filters: bool = False
    ) -> Tuple[List[Dict[str, Any]], int]:
        pushes = []
        for revs, test_datas in test_scheduling.get_test_scheduling_history(
            self.granularity
        ):
            failures = []
            passes = []

            for test_data in test_datas:
                name = test_data["name"]

                if (
                    test_data["is_likely_regression"]
                    or test_data["is_possible_regression"]
                ):
                    failures.append(name)
                else:
                    passes.append(name)

            if apply_filters:
                if self.failures_skip and len(failures) > self.failures_skip:
                    continue

            pushes.append(
                {
                    "revs": revs,
                    "failures": failures,
                    "passes": passes,
                }
            )

        return pushes, math.floor(0.9 * len(pushes))

    # To properly test the performance of our model, we need to split the data
    # according to time: we train on older pushes and evaluate on newer pushes.
    def train_test_split(self, X, y):
        pushes, train_push_len = self.get_pushes(True)
        train_len = sum(
            len(push["failures"]) + len(push["passes"])
            for push in pushes[:train_push_len]
        )
        print(
            f"{train_push_len} pushes in the training set (corresponding to {train_len} push/jobs)"
        )
        return X[:train_len], X[train_len:], y[:train_len], y[train_len:]

    def items_gen(self, classes):
        commit_map = get_commit_map()

        for revs, test_datas in test_scheduling.get_test_scheduling_history(
            self.granularity
        ):
            commits = tuple(
                commit_map.pop(revision) for revision in revs if revision in commit_map
            )
            assert len(commits) > 0

            for test_data in test_datas:
                name = test_data["name"]

                if (revs[0], name) not in classes:
                    continue

                commit_data = commit_features.merge_commits(commits)
                commit_data["test_job"] = test_data
                yield commit_data, classes[(revs[0], name)]

    def get_labels(self):
        classes = {}
        pushes, _ = self.get_pushes(True)

        for push in pushes:
            for name in push["failures"]:
                classes[(push["revs"][0], name)] = 1

            for name in push["passes"]:
                classes[(push["revs"][0], name)] = 0

        print("{} pushes considered".format(len(pushes)))
        print(
            "{} pushes with at least one failure".format(
                sum(1 for push in pushes if len(push["failures"]) > 0)
            )
        )
        print(
            "{} push/jobs failed".format(
                sum(1 for label in classes.values() if label == 1)
            )
        )
        print(
            "{} push/jobs did not fail".format(
                sum(1 for label in classes.values() if label == 0)
            )
        )

        return classes, [0, 1]

    def select_tests(
        self,
        commits: Sequence[repository.CommitDict],
        confidence: float = 0.5,
        push_num: Optional[int] = None,
    ) -> Dict[str, float]:
        commit_data = commit_features.merge_commits(commits)

        past_failures_data = test_scheduling.get_past_failures(self.granularity, True)

        if push_num is None:
            push_num = past_failures_data["push_num"] + 1
        all_runnables = past_failures_data["all_runnables"]

        commit_tests = []
        for data in test_scheduling.generate_data(
            self.granularity,
            past_failures_data,
            commit_data,
            push_num,
            all_runnables,
            tuple(),
            tuple(),
        ):
            commit_test = commit_data.copy()
            commit_test["test_job"] = data
            commit_tests.append(commit_test)

        probs = self.classify(commit_tests, probabilities=True)
        selected_indexes = np.argwhere(probs[:, 1] >= confidence)[:, 0]
        return {
            commit_tests[i]["test_job"]["name"]: math.floor(probs[i, 1] * 100) / 100
            for i in selected_indexes
        }

    def _get_cost(self, config: str) -> int:
        costs = [
            (("build", "opt"), 1),
            (("build", "debug"), 2),
            (("linux1804-64", "opt"), 2),
            (("linux1804-64", "debug"), 3),
            (("windows10", "opt"), 4),
            (("windows10", "debug"), 5),
            (("android-em", "opt"), 6),
            (("android-em", "debug"), 7),
            (("windows7", "opt"), 8),
            (("windows7", "debug"), 9),
            (("mac", "opt"), 10),
            (("mac", "debug"), 11),
            (("asan", "opt"), 12),
            (("asan", "debug"), 13),
            (("linux1804-32", "opt"), 14),
            (("linux1804-32", "debug"), 15),
            (("android-hw", "opt"), 16),
            (("android-hw", "debug"), 17),
            (("tsan", "opt"), 18),
            (("tsan", "debug"), 19),
            (("test-linux1804-64-shippable/opt-*-e10s",), 1),
        ]

        for substrings, cost in reversed(costs):
            if all(s in config for s in substrings):
                return cost

        raise Exception(f"Couldn't find cost for {config}")

    def _generate_equivalence_sets(
        self,
        tasks: Iterable[str],
        min_redundancy_confidence: float,
        load_failing_together: Callable[[str], Dict[str, Tuple[float, float]]],
        assume_redundant: bool,
    ) -> List[Set[str]]:
        # Generate 'equivalence sets', containing all tasks that are redundant with
        # each other.
        groups: List[Set[str]] = []
        task_to_groups: Dict[str, Set[int]] = collections.defaultdict(set)
        incompatible_groups: Dict[str, Set[int]] = collections.defaultdict(set)

        def create_group(task: str) -> None:
            if task in task_to_groups:
                return

            groups.append({task})
            task_to_groups[task] = {len(groups) - 1}

        # Add task1 to all equivalence groups where task2 is present, and likewise for task2.
        # Skip groups which contain tasks that are not redundant with task1.
        def add_to_groups(task1: str, task2: str) -> None:
            found = False

            if task1 in task_to_groups:
                for i in task_to_groups[task1]:
                    if task2 in incompatible_groups and i in incompatible_groups[task2]:
                        continue

                    groups[i].add(task2)
                    task_to_groups[task2].add(i)
                    found = True

            if task2 in task_to_groups:
                for i in task_to_groups[task2]:
                    if task1 in incompatible_groups and i in incompatible_groups[task1]:
                        continue

                    groups[i].add(task1)
                    task_to_groups[task1].add(i)
                    found = True

            # No suitable equivalence group was found for the tasks, create a new one.
            if found:
                return

            group = {task1, task2}
            groups.append(group)
            task_to_groups[task1].add(len(groups) - 1)
            task_to_groups[task2].add(len(groups) - 1)

        def mark_incompatible(task1: str, task2: str) -> None:
            if task1 in task_to_groups:
                incompatible_groups[task2].update(task_to_groups[task1])

            if task2 in task_to_groups:
                incompatible_groups[task1].update(task_to_groups[task2])

        sorted_tasks = sorted(tasks)
        for i, task1 in enumerate(sorted_tasks):
            create_group(task1)

            try:
                failing_together_stats = load_failing_together(task1)
            except KeyError:
                failing_together_stats = {}

            for task2 in sorted_tasks[i + 1 :]:
                try:
                    support, confidence = failing_together_stats[task2]
                except KeyError:
                    if not assume_redundant:
                        confidence = 0.0
                    else:
                        confidence = 1.0

                if confidence >= min_redundancy_confidence:
                    add_to_groups(task1, task2)
                else:
                    mark_incompatible(task1, task2)

        return groups

    def _get_equivalence_sets(self, min_redundancy_confidence: float):
        try:
            with open(
                f"equivalence_sets_{min_redundancy_confidence}.pickle", "rb"
            ) as f:
                return pickle.load(f)
        except FileNotFoundError:
            past_failures_data = test_scheduling.get_past_failures(
                self.granularity, True
            )
            all_runnables = past_failures_data["all_runnables"]

            equivalence_sets = {}
            failing_together = test_scheduling.get_failing_together_db(
                "config_group", True
            )
            all_configs = pickle.loads(failing_together[b"$ALL_CONFIGS$"])
            configs_by_group = pickle.loads(failing_together[b"$CONFIGS_BY_GROUP$"])
            for group in all_runnables:
                key = test_scheduling.failing_together_key(group)
                try:
                    failing_together_stats = pickle.loads(failing_together[key])
                except KeyError:
                    failing_together_stats = {}

                def load_failing_together(
                    config: str,
                ) -> Dict[str, Tuple[float, float]]:
                    return failing_together_stats[config]

                configs = (
                    configs_by_group[group]
                    if group in configs_by_group
                    else all_configs
                )

                equivalence_sets[group] = self._generate_equivalence_sets(
                    configs, min_redundancy_confidence, load_failing_together, True
                )

            with open(
                f"equivalence_sets_{min_redundancy_confidence}.pickle", "wb"
            ) as f:
                pickle.dump(equivalence_sets, f)

            return equivalence_sets

    def _solve_optimization(self, solver: pywraplp.Solver) -> bool:
        # The MIP solver is usually fast (milliseconds). If we hit a weird problem,
        # accept a suboptimal solution after 10 seconds.
        solver.SetTimeLimit(10000)
        status = solver.Solve()

        if status == pywraplp.Solver.INFEASIBLE:
            logger.warning("Optimization problem is infeasible")
            return False
        elif status == pywraplp.Solver.NOT_SOLVED:
            logger.warning("Optimization problem could not be solved in time")
            return False

        return True

    def reduce(
        self,
        tasks: Collection[str],
        min_redundancy_confidence: float,
        assume_redundant: bool = False,
    ) -> Set[str]:
        failing_together = test_scheduling.get_failing_together_db(
            self.granularity, True
        )

        def load_failing_together(task: str) -> Dict[str, Tuple[float, float]]:
            key = test_scheduling.failing_together_key(task)
            return pickle.loads(failing_together[key])

        solver = pywraplp.Solver(
            "select_configs", pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING
        )

        task_vars = {task: solver.BoolVar(task) for task in tasks}

        equivalence_sets = self._generate_equivalence_sets(
            tasks, min_redundancy_confidence, load_failing_together, assume_redundant
        )

        # Create constraints to ensure at least one task from each set of equivalent
        # sets is selected.

        mutually_exclusive = True
        seen = set()
        for equivalence_set in equivalence_sets:
            if any(config in seen for config in equivalence_set):
                mutually_exclusive = False
                break

            seen |= equivalence_set

        for equivalence_set in equivalence_sets:
            sum_constraint = sum(task_vars[task] for task in equivalence_set)
            if mutually_exclusive:
                solver.Add(sum_constraint == 1)
            else:
                solver.Add(sum_constraint >= 1)

        # Choose the best set of tasks that satisfy the constraints with the lowest cost.
        solver.Minimize(
            sum(self._get_cost(task) * task_vars[task] for task in task_vars.keys())
        )

        if self._solve_optimization(solver):
            return {
                task
                for task, task_var in task_vars.items()
                if task_var.solution_value() == 1
            }
        else:
            return set(tasks)

    def select_configs(
        self, groups: Collection[str], min_redundancy_confidence: float
    ) -> Dict[str, List[str]]:
        failing_together = test_scheduling.get_failing_together_db("config_group", True)

        all_configs = pickle.loads(failing_together[b"$ALL_CONFIGS$"])
        all_configs_by_group = pickle.loads(failing_together[b"$CONFIGS_BY_GROUP$"])
        config_costs = {config: self._get_cost(config) for config in all_configs}

        solver = pywraplp.Solver(
            "select_configs", pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING
        )

        config_group_vars = {
            (config, group): solver.BoolVar(f"{group}@{config}")
            for group in groups
            for config in (
                all_configs_by_group[group]
                if group in all_configs_by_group
                else all_configs
            )
        }

        equivalence_sets = self._get_equivalence_sets(min_redundancy_confidence)

        for group in groups:
            # Create constraints to ensure at least one task from each set of equivalent
            # groups is selected.

            mutually_exclusive = True
            seen = set()
            for equivalence_set in equivalence_sets[group]:
                if any(config in seen for config in equivalence_set):
                    mutually_exclusive = False
                    break

                seen |= equivalence_set

            for equivalence_set in equivalence_sets[group]:
                sum_constraint = sum(
                    config_group_vars[(config, group)] for config in equivalence_set
                )
                if mutually_exclusive:
                    solver.Add(sum_constraint == 1)
                else:
                    solver.Add(sum_constraint >= 1)

        # Choose the best set of tasks that satisfy the constraints with the lowest cost.
        solver.Minimize(
            sum(
                config_costs[config] * config_group_vars[(config, group)]
                for config, group in config_group_vars.keys()
            )
        )

        configs_by_group: Dict[str, List[str]] = {}
        for group in groups:
            configs_by_group[group] = []

        if self._solve_optimization(solver):
            for (config, group), config_group_var in config_group_vars.items():
                if config_group_var.solution_value() == 1:
                    configs_by_group[group].append(config)
        else:
            least_cost_config = min(config_costs, key=config_costs.get)
            for group in groups:
                configs_by_group[group].append(least_cost_config)

        return configs_by_group

    def evaluation(self) -> None:
        # Get a test set of pushes on which to test the model.
        pushes, train_push_len = self.get_pushes(False)

        # To evaluate the model with reductions enabled, we need to regenerate the failing together DB, using
        # only failure data from the training pushes (otherwise, we'd leak training information into the test
        # set).
        print("Generate failing together DB (restricted to training pushes)")
        push_data_iter, push_data_count, _ = test_scheduling.get_push_data(
            "label" if self.granularity == "label" else "config_group"
        )
        test_scheduling.generate_failing_together_probabilities(
            "label" if self.granularity == "label" else "config_group",
            push_data_iter(),
            push_data_count,
            pushes[train_push_len - 1]["revs"][0],
        )

        test_pushes_list = pushes[train_push_len:]

        all_tasks = reduce(
            lambda x, y: x | y,
            (
                set(push["failures"]) | set(push["passes"])
                for push in test_pushes_list[-28:]
            ),
        )

        all_revs = set(sum((push["revs"] for push in test_pushes_list), []))

        test_pushes_failures = sum(
            1 for push in test_pushes_list if len(push["failures"]) > 0
        )

        test_pushes = {push["revs"][0]: push for push in test_pushes_list}

        print(
            f"Testing on {len(test_pushes)} ({test_pushes_failures} with failures) out of {len(pushes)}. {len(all_tasks)} schedulable tasks."
        )

        del pushes

        commit_map = get_commit_map(all_revs)

        past_failures_data = test_scheduling.get_past_failures(self.granularity, True)
        last_push_num = past_failures_data["push_num"]
        past_failures_data.close()

        # Select tests for all the pushes in the test set.
        for i, push in enumerate(tqdm(test_pushes.values())):
            commits = tuple(
                commit_map.pop(revision)
                for revision in push["revs"]
                if revision in commit_map
            )
            if len(commits) == 0:
                push["all_possibly_selected"] = {}
                continue

            push_num = last_push_num - (len(test_pushes) - (i + 1))

            # Note: we subtract 100 to the push number to make sure we don't use
            # past failure data for the push itself.
            # The number 100 comes from the fact that in the past failure data
            # generation we store past failures in batches of 100 pushes.
            push["all_possibly_selected"] = self.select_tests(
                commits, 0.5, push_num - 100
            )

        def do_eval(
            executor: concurrent.futures.ProcessPoolExecutor,
            confidence_threshold: float,
            reduction: Optional[float],
            cap: Optional[int],
            minimum: Optional[int],
        ) -> None:
            futures: Dict[concurrent.futures.Future, Dict[str, Any]] = {}
            for push in test_pushes.values():
                futures[
                    executor.submit(
                        eval_apply_transforms,
                        self,
                        push,
                        confidence_threshold,
                        reduction,
                        cap,
                        minimum,
                    )
                ] = push

            for future in concurrent.futures.as_completed(futures):
                exc = future.exception()
                if exc is not None:
                    print(
                        "Exception {} while running {}".format(
                            exc, futures[future]["revs"][0]
                        )
                    )
                    for f in futures:
                        f.cancel()

                push = futures[future]
                selected, number_configs = future.result()

                if reduction is not None and self.granularity == "group":
                    push["number_configs"] = number_configs

                caught = selected & set(push["failures"])

                push["number_scheduled"] = len(selected)
                push["caught_one"] = (
                    len(caught) > 0 if len(push["failures"]) != 0 else None
                )
                push["some_didnt_run"] = (
                    not selected.issubset(set(push["passes"]) | set(push["failures"])),
                )
                push["caught_percentage"] = (
                    len(caught) / len(push["failures"])
                    if len(push["failures"]) != 0
                    else None
                )

            min_scheduled = min(
                result["number_scheduled"] for result in test_pushes.values()
            )
            max_scheduled = max(
                result["number_scheduled"] for result in test_pushes.values()
            )
            average_scheduled = statistics.mean(
                result["number_scheduled"] for result in test_pushes.values()
            )
            num_failing_pushes = sum(
                1 for result in test_pushes.values() if result["caught_one"] is not None
            )
            num_caught_one = sum(
                1 for result in test_pushes.values() if result["caught_one"]
            )
            num_caught_one_or_some_didnt_run = sum(
                1
                for result in test_pushes.values()
                if result["caught_one"]
                or (result["caught_one"] is not None and result["some_didnt_run"])
            )
            percentage_caught_one = 100 * num_caught_one / num_failing_pushes
            percentage_caught_one_or_some_didnt_run = (
                100 * num_caught_one_or_some_didnt_run / num_failing_pushes
            )
            average_caught_percentage = 100 * statistics.mean(
                result["caught_percentage"]
                for result in test_pushes.values()
                if result["caught_percentage"] is not None
            )

            reduction_str = (
                f"enabled at {reduction * 100}%"
                if reduction is not None
                else "disabled"
            )

            message = f"For confidence threshold {confidence_threshold}, with reduction {reduction_str}, cap at {cap}, and minimum at {minimum}: scheduled {average_scheduled} tasks on average (min {min_scheduled}, max {max_scheduled}). In {percentage_caught_one}% of pushes we caught at least one failure ({percentage_caught_one_or_some_didnt_run}% ignoring misses when some of our selected tasks didn't run). On average, we caught {average_caught_percentage}% of all seen failures."

            if reduction is not None and self.granularity == "group":
                average_configs = statistics.mean(
                    result["number_configs"] for result in test_pushes.values()
                )
                message += f" On average, we selected {average_configs} configs."

            print(message)

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=utils.get_physical_cpu_count()
        ) as executor:
            scenarios = [
                (None, None, None),
                (10, None, None),
                (None, 300, None),
                (None, None, 0.9),
                (None, None, 1.0),
            ]
            for minimum, cap, reduction in scenarios:
                # Pre-generate equivalence sets, so when we run the config selection in multiple processes
                # we don't risk concurrent writes to the equivalence sets file.
                if reduction is not None and self.granularity == "group":
                    self._get_equivalence_sets(reduction)

                for confidence_threshold in [0.5, 0.7, 0.8, 0.85, 0.9, 0.95]:
                    do_eval(executor, confidence_threshold, reduction, cap, minimum)

    def get_feature_names(self):
        return self.extraction_pipeline.named_steps["union"].get_feature_names()


class TestLabelSelectModel(TestSelectModel):
    def __init__(self, lemmatization=False):
        TestSelectModel.__init__(self, lemmatization, "label", failures_skip=60)


class TestGroupSelectModel(TestSelectModel):
    def __init__(self, lemmatization=False):
        TestSelectModel.__init__(self, lemmatization, "group")


class TestConfigGroupSelectModel(TestSelectModel):
    def __init__(self, lemmatization=False):
        TestSelectModel.__init__(self, lemmatization, "config_group")


def eval_apply_transforms(model, push, confidence_threshold, reduction, cap, minimum):
    number_configs = None

    selected = set(
        name
        for name, confidence in push["all_possibly_selected"].items()
        if confidence >= confidence_threshold
    )

    if reduction is not None:
        if model.granularity == "label":
            selected = model.reduce(selected, reduction)
        elif model.granularity == "group":
            number_configs = len(
                set(
                    sum(
                        model.select_configs(selected, reduction).values(),
                        [],
                    )
                )
            )

    if minimum is not None and len(selected) < minimum:
        remaining = [
            (name, confidence)
            for name, confidence in push["all_possibly_selected"].items()
            if name not in selected
        ]
        selected.update(
            name
            for name, _ in sorted(remaining, key=lambda x: -x[1])[
                : minimum - len(selected)
            ]
        )

    if cap is not None and len(selected) > cap:
        selected = set(
            sorted(
                (
                    (name, confidence)
                    for name, confidence in push["all_possibly_selected"].items()
                    if name in selected
                ),
                key=lambda x: x[1],
                reverse=True,
            )[:cap]
        )

    return selected, number_configs



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import gzip
import io
import logging
import os
import pickle
from contextlib import contextmanager
from urllib.parse import urljoin

import orjson
import requests
import zstandard

from bugbug import utils

DATABASES = {}

logger = logging.getLogger(__name__)


def register(path, url, version, support_files=[]):
    DATABASES[path] = {"url": url, "version": version, "support_files": support_files}

    # Create DB parent directory.
    os.makedirs(os.path.abspath(os.path.dirname(path)), exist_ok=True)

    if not os.path.exists(f"{path}.version"):
        with open(f"{path}.version", "w") as f:
            f.write(str(version))


def is_registered(path: str) -> bool:
    return path in DATABASES


def exists(path):
    return os.path.exists(path)


def is_old_schema(path):
    url = urljoin(DATABASES[path]["url"], f"{os.path.basename(path)}.version")
    r = requests.get(url)

    if not r.ok:
        logger.info(f"Version file is not yet available to download for {path}")
        return True

    prev_version = int(r.text)

    return DATABASES[path]["version"] > prev_version


def download_support_file(path, file_name, extract=True):
    # If a DB with the current schema is not available yet, we can't download.
    if is_old_schema(path):
        return False

    try:
        url = urljoin(DATABASES[path]["url"], file_name)
        path = os.path.join(os.path.dirname(path), file_name)

        logger.info(f"Downloading {url} to {path}")
        updated = utils.download_check_etag(url, path)

        if extract and updated and path.endswith(".zst"):
            utils.extract_file(path)
            os.remove(path)

        return True
    except requests.exceptions.HTTPError:
        logger.info(
            f"{file_name} is not yet available to download for {path}", exc_info=True
        )
        return False


# Download and extract databases.
def download(path, support_files_too=False, extract=True):
    # If a DB with the current schema is not available yet, we can't download.
    if is_old_schema(path):
        return False

    zst_path = f"{path}.zst"

    url = DATABASES[path]["url"]
    try:
        logger.info(f"Downloading {url} to {zst_path}")
        updated = utils.download_check_etag(url, zst_path)

        if extract and updated:
            utils.extract_file(zst_path)
            os.remove(zst_path)

        successful = True
        if support_files_too:
            for support_file in DATABASES[path]["support_files"]:
                successful |= download_support_file(path, support_file, extract)

        return successful
    except requests.exceptions.HTTPError:
        logger.info(f"{url} is not yet available to download", exc_info=True)
        return False


def upload(path):
    support_files_paths = [
        os.path.join(os.path.dirname(path), support_file_path)
        for support_file_path in DATABASES[path]["support_files"]
    ]
    utils.upload_s3([f"{path}.zst", f"{path}.version"] + support_files_paths)


def last_modified(path):
    url = DATABASES[path]["url"]
    last_modified = utils.get_last_modified(url)

    if last_modified is None:
        raise Exception("Last-Modified is not available")

    return last_modified


class Store:
    def __init__(self, fh):
        self.fh = fh


class JSONStore(Store):
    def write(self, elems):
        for elem in elems:
            self.fh.write(orjson.dumps(elem) + b"\n")

    def read(self):
        for line in io.TextIOWrapper(self.fh, encoding="utf-8"):
            yield orjson.loads(line)


class PickleStore(Store):
    def write(self, elems):
        for elem in elems:
            self.fh.write(pickle.dumps(elem))

    def read(self):
        try:
            while True:
                yield pickle.load(self.fh)
        except EOFError:
            pass


COMPRESSION_FORMATS = ["gz", "zstd"]
SERIALIZATION_FORMATS = {"json": JSONStore, "pickle": PickleStore}


@contextmanager
def _db_open(path, mode):
    parts = str(path).split(".")
    assert len(parts) > 1, "Extension needed to figure out serialization format"
    if len(parts) == 2:
        db_format = parts[-1]
        compression = None
    else:
        db_format = parts[-2]
        compression = parts[-1]

    assert compression is None or compression in COMPRESSION_FORMATS
    assert db_format in SERIALIZATION_FORMATS

    store_constructor = SERIALIZATION_FORMATS[db_format]

    if compression == "gz":
        with gzip.GzipFile(path, mode) as f:
            yield store_constructor(f)
    elif compression == "zstd":
        if "w" in mode or "a" in mode:
            cctx = zstandard.ZstdCompressor()
            with open(path, mode) as f:
                with cctx.stream_writer(f) as writer:
                    yield store_constructor(writer)
        else:
            dctx = zstandard.ZstdDecompressor()
            with open(path, mode) as f:
                with dctx.stream_reader(f) as reader:
                    yield store_constructor(reader)
    else:
        with open(path, mode) as f:
            yield store_constructor(f)


def read(path):
    assert path in DATABASES

    if not os.path.exists(path):
        return ()

    with _db_open(path, "rb") as store:
        for elem in store.read():
            yield elem


def write(path, elems):
    assert path in DATABASES

    with _db_open(path, "wb") as store:
        store.write(elems)


def append(path, elems):
    assert path in DATABASES

    with _db_open(path, "ab") as store:
        store.write(elems)


def delete(path, match):
    assert path in DATABASES

    dirname, basename = os.path.split(path)
    new_path = os.path.join(dirname, f"new_{basename}")

    def matching_elems(store):
        for elem in store.read():
            if not match(elem):
                yield elem

    try:
        with _db_open(path, "rb") as rstore:
            with _db_open(new_path, "wb") as wstore:
                wstore.write(matching_elems(rstore))
    except FileNotFoundError:
        return

    os.unlink(path)
    os.rename(new_path, path)



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
from typing import Collection, Iterator, NewType, Optional

from libmozdata.phabricator import PhabricatorAPI
from tqdm import tqdm

from bugbug import db

logger = logging.getLogger(__name__)

RevisionDict = NewType("RevisionDict", dict)

REVISIONS_DB = "data/revisions.json"
db.register(
    REVISIONS_DB,
    "https://community-tc.services.mozilla.com/api/index/v1/task/project.bugbug.data_revisions.latest/artifacts/public/revisions.json.zst",
    1,
)

PHABRICATOR_API = None

TESTING_PROJECTS = {
    "PHID-PROJ-h7y4cs7m2o67iczw62pp": "testing-approved",
    "PHID-PROJ-e4fcjngxcws3egiecv3r": "testing-exception-elsewhere",
    "PHID-PROJ-iciyosoekrczpf2a4emw": "testing-exception-other",
    "PHID-PROJ-zjipshabawolpkllehvg": "testing-exception-ui",
    "PHID-PROJ-cspmf33ku3kjaqtuvs7g": "testing-exception-unchanged",
}


def get_revisions() -> Iterator[RevisionDict]:
    yield from db.read(REVISIONS_DB)


def set_api_key(url: str, api_key: str) -> None:
    global PHABRICATOR_API
    PHABRICATOR_API = PhabricatorAPI(api_key, url)


def get(rev_ids: Collection[int]) -> Collection[RevisionDict]:
    assert PHABRICATOR_API is not None

    out = PHABRICATOR_API.request(
        "differential.revision.search",
        constraints={
            "ids": rev_ids,
        },
        attachments={"projects": True},
    )

    return out["data"]


def download_revisions(rev_ids: Collection[int]) -> None:
    old_rev_count = 0
    new_rev_ids = set(int(rev_id) for rev_id in rev_ids)
    for rev in get_revisions():
        old_rev_count += 1
        if rev["id"] in new_rev_ids:
            new_rev_ids.remove(rev["id"])

    print(f"Loaded {old_rev_count} revisions.")

    new_rev_ids_list = sorted(list(new_rev_ids))
    rev_ids_groups = (
        new_rev_ids_list[i : i + 100] for i in range(0, len(new_rev_ids_list), 100)
    )

    with tqdm(total=len(new_rev_ids)) as progress_bar:
        for rev_ids_group in rev_ids_groups:
            revisions = get(rev_ids_group)

            progress_bar.update(len(rev_ids_group))

            db.append(REVISIONS_DB, revisions)


def get_testing_project(rev: RevisionDict) -> Optional[str]:
    testing_projects = [
        TESTING_PROJECTS[projectPHID]
        for projectPHID in rev["attachments"]["projects"]["projectPHIDs"]
        if projectPHID in TESTING_PROJECTS
    ]

    if len(testing_projects) > 1:
        logger.warning("Revision D{} has more than one testing tag.".format(rev["id"]))

    if len(testing_projects) == 0:
        return None

    return testing_projects[-1]
